#include "common.h"
/* $Header: edit.c,v 2.1 89/09/08 11:51:15 gary Locked $ */
#include "parser.h"
#include "ast.h"
#include "manager.h"
#include "scanner.h"
#include "micro.h"
#include <ctype.h>

#define MAXBUFF 1000
extern	void rev( STR X STR ),
	rev_copy( STR X STR ),
	readstr(STR);

/* maintain last MAXERRBUFF chars read for error messages */
#define MAXERRBUFF 200
static char errbuff[MAXERRBUFF];


static int in_err = 0;
static Bool errfull = false;


extern FILE *CONin;		       /* Console input file  */


/* Print the last lines of input */
void    Show_Error()
{
    int     i;

    mcputc('\n');
    if (errfull) {
	for (i = in_err; i < MAXERRBUFF; i++) {
	    mcprintf("%c", errbuff[i],0);
	}
    }
    for (i = 0; i < in_err; i++) {
	mcprintf("%c", errbuff[i],0);
    }
    mcputc('\n');
}

      static	char	buff[ MAXBUFF ] = {0};
      static	char *	in_buff = buff;
      static	char *	out_buff;
      static	char	fstate = 'N';
      /*	Fetch State
	      (N) normal	read from yyin
	      (E) edit	read from out_buff, then go to normal
      */    


      static	char	bstate = 'N';
      /*	Buffer state
	      (N) normal	buffer
	      (!) special	don't buffer, revert to normal on '\n'
      */    


/* Clear error buffer */
void    Clear()
{
    in_buff = buff;
    in_err = 0;
    errfull = false;
    func_level = 0;
    Func_List  = Nil;
}


Bool    last_nl = false;	       /* was last char echoed a newline */
char    last_caught = EOS;	       /* last char returned from catch */

/* Print newline if echoing input */
void    New_Line(NONE);
void    New_Line()
{
    if (!last_nl && echo_mode)
	mcputc('\n');
    last_nl = true;
}

void    reset_line()
{
    while( last_caught != '\n' ) catch(); /* keep reading */ 
}


int immediate_char(){
     int c;
     if (direct_input) {
	c = getc(yyin);
    } else if (yyin == CONin || yyin == stdin) {
#ifdef Screen
	c = Egetc();
#else
	c = ile_getc();
#endif
    } else {
	c = getc(yyin);
    }
    if (c == '\r')
	c = '\n';
    return c;
}


/*
 * supplies characters for input.
 * if !edit, may take characters from editting buffer, "!clear", or yyin
 *
 * also buffers input chars
 */
#define Max_Line 200
static char catch_line[Max_Line+1];
static int  c_next=0, c_len=0;
int     catch()
{
    int     c;

    /* keep buffer full */
    while( c_next >= c_len ) {
	c_next = 0;
	do {
	    if( edit_flag ) {
		switch( yyin==stdin ? fstate : 'N' ) {
		case 'N':
		    c = getc(yyin);
		    break;

		case 'E':	c = *(out_buff++);
		    if( c == EOS ) {
			fstate = 'N';
			silent = false;  /* turn prompts back on */
			c = '\n';
		    }
		    break;

		default:	assert(0);
		}
	    } else {
	       c = immediate_char();
	    }

	    catch_line[c_next++] = c;

	} while( c != EOF && c != '\n' && c_next < Max_Line );
	c_len = c_next;
	catch_line[c_len] = EOS;
	c_next = 0;
	if(echo_mode) {
	    char * p = catch_line;
	    for( ; *p != EOS ; p++ ) {
		if( *p != (char)EOF ) mcputc(*p);
	    }
	}
	last_nl = (c == '\n');
    }
    /* c = next char */
    c = catch_line[c_next++];
    if( c==(char)EOF ) c = EOF;

    if (c != EOF) {
	/* buffer c */
	extern FILE *record;

	if (yyin == stdin && record != NULL) {
	    fputc(c, record);
	    if (c == '\n')
		fflush(record);
	}
	errbuff[in_err] = c;
	in_err = (in_err + 1) % MAXERRBUFF;
	if (in_err == 0)
	    errfull = true;

	if( edit_flag )switch( bstate ) {
	    case 'N':	if( c == '!' ) {
			    bstate = '!';
			} else if( yyin != stdin ) {
			    /* do nothing */
			} else if( in_buff==buff && isspace(c) ) {
			    /* do nothing */
			} else {
			    *(in_buff++)	= c;
			    if( in_buff == buff + MAXBUFF ) {
				printf("!Input buffer overflow\n");
				in_buff = buff;
			    }
			}
			break;

	    case '!':	if( c == '\n' )	 {
			    bstate = 'N';
			}
			break;

	    default:	assert(0);
	}
    }
    if( c == '\n' ) Source_Flush();
    
    last_caught = c;
    return c;
}

char * locate( STR );

static char * locate(s) char * s; {
    char * possible;
    int len = strlen(s);
    extern char * index();

    possible = buff-1;
    do {
	possible = index(possible+1, s[0]);
    } while( possible != CNULL && strncmp(possible, s, len) != 0 );

    return possible;
}

#define MAXPAT	100
int Edit() {
    char	pattern[MAXPAT];
    char	replacement[MAXPAT];
    char *	where;
    int		buff_len, pat_len, repl_len;

    *in_buff = EOS;
    while( true ) {
	buff_len = strlen(buff);
	/* trim trailing white space */
	    while( isspace(buff[buff_len-1]) ){
		buff_len--;
	    }
	    buff[buff_len] = EOS;
	printf("%s\n", buff );
	printf("!Enter string to replace (blank line to execute)\n");

	readstr(pattern);
	pat_len = strlen(pattern);
	if( pat_len == 0 ) break;

	/* find pattern */
	    where = locate(pattern);
	    if( where == CNULL ) {
		printf("!String not found.\n" );
		continue;
	    }
	
	printf("!Enter replacement\n");
	readstr(replacement);
	repl_len = strlen(replacement);

	/* replace pattern with replacement */
	    if( buff_len - pat_len + repl_len < MAXBUFF ) {
		int front_len	= buff - where;
		int tail_len	= buff_len - front_len - pat_len;

		/* buff = front pattern tail */
		rev( where, buff + buff_len - 1 );
		/* buff = front R(tail) R(pattern) */
		rev_copy( replacement,
			  buff + front_len + tail_len );
		/* buff = front R(tail) R(replacement) */
		rev( where,
		     buff + front_len + tail_len + repl_len - 1 );
	    } else {
		printf("!Replacement would overflow buffer.  Ignored.\n");
	    }
    }
    fstate = 'E';
    silent = true;	/* turn off prompts */
    out_buff = buff;
    editting = true;
    in_buff  = buff;
    return(CLEAR);
}
/**/
static void readstr(s) char * s; {
    int i, c;

    i = 0;
    while( true ) {
	c = getchar();
	if( c == '\n' )		break;
	if( i<MAXPAT )		s[i++] = c;
    }

    if( i == MAXPAT ) {
	printf("!Input too long\n" );
	s[MAXPAT-1] = EOS;
    } else {
	s[i] = EOS;
    }
}


static void rev(start, stop) char * start, * stop; {
    while( start < stop ) {
	char tmp = *start;
	*start = *stop;
	*stop  = tmp;
	start++;
	stop--;
    }
}

static void rev_copy(from, to) char * from, * to; {
    to = to + strlen(from);
    *to = EOS;
    to --;
    while( *from ) *(to --) = *(from ++);
}


/* Indicate that you are done with the current source line
 * Link in all the statements that started on this line
 */
void Source_Flush()
{
    ENTER;
    Reg(line);
    AST_Ptr t;

    if (store_source) {
	R(line) = New_String(catch_line);
    } else {
	R(line) = ANULL;
    }
    if( Func_List != Nil ) {
	assert(Kind(Func_Tail)==List);
	Next(Func_Tail) = mcCons(R(line), Nil);
	Func_Tail = Next(Func_Tail);
    }

    while (Source_List != ANULL) {
	t = Source_List;
	Source_List = ast_1(Source_List);
	ast_1(t) = R(line);
    }

    LEAVE;
}
