#include "common.h"
/* $Header: op_arith.c,v 2.1 89/09/08 11:53:01 gary Locked $ */
#include "instr.h"
#include "op.h"
#include "manager.h"
#include "micro.h"
#include "emitter.h"
/* Because these are ISETL operations, the arguments are on the ISETL
 * stack, not passed as C arguments.  Comments refer to the arguments
 * on the stack.  They are correctly accessed by the ENTER macros.
 * See op.h for an explanation of ENTER, LEAVE, R(X_Op).
 * See instr.h for further description of ISETL operations.
 */


/* (a,b):*  a+b */
void    opPlus()
{
    BIN_ENTER;

    if (Kind(R(Left_Op)) == Kind(R(Right_Op))) {
	switch (Kind(R(Left_Op))) {
	  case Integer:
	    R(ans) = New_Integer((WORKING) Int_Value(R(Left_Op))
				 + Int_Value(R(Right_Op)));
	    break;

	  case Bignum:
	    R(ans) = mpAdd(R(Left_Op), R(Right_Op));
	    break;

	  case Real:
	    R(ans) = New_Real(Real_Value(R(Left_Op))
			      + Real_Value(R(Right_Op)));
	    break;

	  case Set:
	    R(ans) = mcUnion(Set_Value(R(Left_Op)),
			     Set_Value(R(Right_Op)));
	    R(ans) = New_Set(R(ans));
	    break;

	  case String:
	    R(ans) = mcStr_Concat(R(Left_Op), R(Right_Op));
	    break;

	  case Tuple:
	    R(ans) = mcCat(Tuple_Value(R(Left_Op)),
			   Tuple_Value(R(Right_Op)));
	    R(ans) = New_Tuple(R(ans), Tuple_Origin(R(Left_Op)));
	    break;

	  default:
	    BAD_ARGS("+", R(Left_Op), R(Right_Op));
	}
    } else {
	/* mixed operations */
	switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
	  case Pair(Integer, Bignum):
	  case Pair(Bignum, Integer):{
		R(Left_Op) = mpForce(R(Left_Op));
		R(Right_Op) = mpForce(R(Right_Op));
		R(ans) = mpAdd(R(Left_Op), R(Right_Op));
	    } break;

	  case Pair(Integer, Real):
	  case Pair(Real, Integer):
	  case Pair(Bignum, Real):
	  case Pair(Real, Bignum):
	    R(Left_Op) = mpFloat(R(Left_Op));
	    R(Right_Op) = mpFloat(R(Right_Op));
	    R(ans) = New_Real(Real_Value(R(Left_Op))
			      + Real_Value(R(Right_Op)));
	    break;

	  default:
	    BAD_ARGS("+", R(Left_Op), R(Right_Op));
	}
    }
    BIN_LEAVE;
}



/* (a,b):*  a-b */
void    opMinus()
{
    BIN_ENTER;

    if (Kind(R(Left_Op)) == Kind(R(Right_Op))) {
	switch (Kind(R(Left_Op))) {
	  case Integer:
	    R(ans) = New_Integer((WORKING) Int_Value(R(Left_Op))
				 - Int_Value(R(Right_Op)));
	    break;

	  case Bignum:
	    R(ans) = mpSub(R(Left_Op), R(Right_Op));
	    break;

	  case Real:
	    R(ans) = New_Real(Real_Value(R(Left_Op))
			      - Real_Value(R(Right_Op)));
	    break;

	  case Set:
	    R(ans) = mcDiff(Set_Value(R(Left_Op)),
			    Set_Value(R(Right_Op)));
	    R(ans) = New_Set(R(ans));
	    break;
	  default:
	    BAD_ARGS("-", R(Left_Op), R(Right_Op));
	}
    } else {
	/* mixed operations */
	switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
	  case Pair(Integer, Bignum):
	  case Pair(Bignum, Integer):{
		R(Left_Op) = mpForce(R(Left_Op));
		R(Right_Op) = mpForce(R(Right_Op));
		R(ans) = mpSub(R(Left_Op), R(Right_Op));
	    } break;

	  case Pair(Integer, Real):
	  case Pair(Real, Integer):
	  case Pair(Bignum, Real):
	  case Pair(Real, Bignum):
	    R(Left_Op) = mpFloat(R(Left_Op));
	    R(Right_Op) = mpFloat(R(Right_Op));
	    R(ans) = New_Real(Real_Value(R(Left_Op))
			      - Real_Value(R(Right_Op)));
	    break;

	  default:
	    BAD_ARGS("-", R(Left_Op), R(Right_Op));
	}
    }
    BIN_LEAVE;
}


void    opNegate()
{
    UNARY_ENTER;
    switch (Kind(R(Only_Op))) {
      case Integer:
	R(ans) = New_Block(Integer, 0);
	Int_Value(R(ans)) = -Int_Value(R(Only_Op));
	break;

      case Bignum:
	R(ans) = New_MP(Length(R(Only_Op)));
	mpcpy(R(ans), R(Only_Op));
	Sign(R(ans)) = -Sign(R(ans));
	break;

      case Real:
	R(ans) = New_Real(-Real_Value(R(Only_Op)));
	break;

      default:
	BAD_ARGS("-", NULL, R(Only_Op));
    }
    UNARY_LEAVE;
}



/* (a,b):*  a*b */
void    opTimes()
{
    BIN_ENTER;

    if (Kind(R(Left_Op)) == Kind(R(Right_Op))) {
	switch (Kind(R(Left_Op))) {
	  case Integer:{
		WORKING result;

		result = ((WORKING) Int_Value(R(Left_Op)))
		    * Int_Value(R(Right_Op));
		R(ans) = New_Integer(result);
	    } break;

	  case Bignum:
	    R(ans) = mpTimes(R(Left_Op), R(Right_Op));
	    break;

	  case Real:
	    R(ans) = New_Real(Real_Value(R(Left_Op))
			      * Real_Value(R(Right_Op)));
	    break;

	  case Set:
	    R(ans) = mcInter(Set_Value(R(Left_Op)),
			     Set_Value(R(Right_Op)));
	    R(ans) = New_Set(R(ans));
	    break;
	  default:
	    BAD_ARGS("*", R(Left_Op), R(Right_Op));
	}
    } else {
	/* mixed operations */
	switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
	  case Pair(Integer, Bignum):
	  case Pair(Bignum, Integer):{
		R(Left_Op) = mpForce(R(Left_Op));
		R(Right_Op) = mpForce(R(Right_Op));
		R(ans) = mpTimes(R(Left_Op), R(Right_Op));
	    } break;

	  case Pair(Integer, Real):
	  case Pair(Real, Integer):
	  case Pair(Bignum, Real):
	  case Pair(Real, Bignum):
	    R(Left_Op) = mpFloat(R(Left_Op));
	    R(Right_Op) = mpFloat(R(Right_Op));
	    R(ans) = New_Real(Real_Value(R(Left_Op))
			      * Real_Value(R(Right_Op)));
	    break;

	  case Pair(Integer, String):
	    R(ans) = mcStr_Rep(Int_Value(R(Left_Op)),
			       R(Right_Op));
	    break;

	  case Pair(String, Integer):
	    R(ans) = mcStr_Rep(Int_Value(R(Right_Op)),
			       R(Left_Op));
	    break;

	  case Pair(Integer, Tuple):
	    R(ans) = mcRep(Int_Value(R(Left_Op)),
			   Tuple_Value(R(Right_Op)));
	    R(ans) = New_Tuple(R(ans), Tuple_Origin(R(Right_Op)));
	    break;

	  case Pair(Tuple, Integer):
	    R(ans) = mcRep(Int_Value(R(Right_Op)),
			   Tuple_Value(R(Left_Op)));
	    R(ans) = New_Tuple(R(ans), Tuple_Origin(R(Left_Op)));
	    break;

	  default:
	    BAD_ARGS("*", R(Left_Op), R(Right_Op));
	}
    }

    BIN_LEAVE;
}



/*
 * Defined only for non-zero divisors, result is always non-negative.
 * See opDiv for accommodations so that (a div b)*b + (a mod b) = a
 * and (a div -b) = -(a div b)
 */
void    opMod()
{
    C_Int   a,
            b,
	    r;

    BIN_ENTER;

    switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
      case Pair(Integer, Integer):
	a = Int_Value(R(Left_Op));
	b = Int_Value(R(Right_Op));

	if (b < 0) {
	    b = -b;
	} else if (b == 0) {
	    BAD_ARGS("mod", R(Left_Op), R(Right_Op));
	}
	r = a % b;
	if (r < 0)
	    r = r + b;
	R(ans) = New_Integer((WORKING) r);
	break;

      case Pair(Bignum, Bignum):{
	    Reg     (a2);

	    if (Length(R(Right_Op)) == 0) {
		BAD_ARGS("mod", R(Left_Op), R(Right_Op));
	    }
	    mpDivMod(R(Left_Op), R(Right_Op), &R(a2), &R(ans));
	    /* handles negative b correctly */
	} break;

      case Pair(Integer, Bignum):
      case Pair(Bignum, Integer):{
	    Reg     (a2);

	    R(Left_Op) = mpForce(R(Left_Op));
	    R(Right_Op) = mpForce(R(Right_Op));
	    if (Length(R(Right_Op)) == 0) {
		BAD_ARGS("mod", R(Left_Op), R(Right_Op));
	    }
	    mpDivMod(R(Left_Op), R(Right_Op), &R(a2), &R(ans));
	} break;
      default:
	BAD_ARGS("mod", R(Left_Op), R(Right_Op));
    }

    BIN_LEAVE;
}



void    opDiv()
{
    C_Int   a,
            b,
            q,
            s;

    BIN_ENTER;

    switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
      case Pair(Integer, Integer):
	a = Int_Value(R(Left_Op));
	b = Int_Value(R(Right_Op));

	if (b == 0) {
	    RT_ERROR("Divide by zero");
	} else if (b > 0) {
	    s = 1;
	} else {
	    s = -1;
	    b = -b;
	}

	q = a / b;
	/* if a mod b < 0 add |b| to make it positive,	 */
	/* therefore subtract 1 from quotient	 */
	if (a % b < 0)
	    q = q - 1;

	q = q * s;

	R(ans) = New_Integer((WORKING) q);
	break;

      case Pair(Bignum, Bignum):{
	    Reg     (a2);

	    if (Length(R(Right_Op)) == 0) {
		RT_ERROR("Divide by zero");
	    }
	    mpDivMod(R(Left_Op), R(Right_Op), &R(ans), &R(a2));
	} break;

      case Pair(Integer, Bignum):
      case Pair(Bignum, Integer):{
	    Reg     (a2);

	    R(Left_Op) = mpForce(R(Left_Op));
	    R(Right_Op) = mpForce(R(Right_Op));
	    if (Length(R(Right_Op)) == 0) {
		RT_ERROR("Divide by zero");
	    }
	    mpDivMod(R(Left_Op), R(Right_Op), &R(ans), &R(a2));
	} break;

      default:
	BAD_ARGS("div", R(Left_Op), R(Right_Op));
    }

    BIN_LEAVE;
}


/* (a,b):*  a/b */
void    opSlash()
{
    C_Real  a,
            b;

    BIN_ENTER;

    if (Kind(R(Left_Op)) == Kind(R(Right_Op))) {
	switch (Kind(R(Left_Op))) {
	  case Integer:
	  case Real:
	  case Bignum:
	    R(Left_Op) = mpFloat(R(Left_Op));
	    R(Right_Op) = mpFloat(R(Right_Op));
	    break;

	  default:
	    BAD_ARGS("/", R(Left_Op), R(Right_Op));
	}
    } else {
	/* mixed operations */
	switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
	  case Pair(Integer, Bignum):
	  case Pair(Bignum, Integer):
	  case Pair(Integer, Real):
	  case Pair(Real, Integer):
	  case Pair(Bignum, Real):
	  case Pair(Real, Bignum):
	    R(Left_Op) = mpFloat(R(Left_Op));
	    R(Right_Op) = mpFloat(R(Right_Op));
	    break;

	  default:
	    BAD_ARGS("/", R(Left_Op), R(Right_Op));
	}
    }
    a = Real_Value(R(Left_Op));
    b = Real_Value(R(Right_Op));

    if (b == 0.0) {
	RT_ERROR("Divide by zero");
    }
    R(ans) = New_Real(a / b);

    BIN_LEAVE;
}


/* (B,Ex): B ** Ex */
void    opExp()
{
    BIN_ENTER;
    switch (Pair(Kind(R(Left_Op)), Kind(R(Right_Op)))) {
      case Pair(Bignum, Bignum):
      case Pair(Bignum, Integer):
      case Pair(Integer, Bignum):
      case Pair(Integer, Integer):{
	    Reg     (x);
	    Reg     (y);
	    Reg     (t);
	    Reg     (r);
	    int     s = 0;

	    R(x) = mpForce(R(Left_Op));
	    R(y) = mpForce(R(Right_Op));


	    if (Sign(R(y)) < 0 || (Length(R(x)) == 0 && Length(R(y)) == 0)) {
		R(ans) = OM;
	    } else {
		/* Left**Right == ans*(x**y) */
		R(ans) = BigOne;
		while (true) {
		    mpUDivMod(R(y), BigTwo, &R(t), &R(r));
		    mpNorm(R(r));
		    mpNorm(R(t));
		    if (s == 0) {
			if (Length(R(r)) == 0)
			    s = 1;
			else
			    s = Sign(R(x));
		    }
		    R(y) = R(t);
		    if (Length(R(r)) != 0) {
			R(ans) = mpUTimes(R(ans), R(x));
			mpNorm(R(ans));
		    }
		    if (Length(R(y)) == 0)
			break;

		    R(x) = mpUTimes(R(x), R(x));
		    mpNorm(R(x));
		}
		if (s == 0)
		    s = 1;
		Sign(R(ans)) = s;
		R(ans) = mpNorm(R(ans));
	    }
	} break;

      case Pair(Integer, Real):
      case Pair(Real, Integer):
      case Pair(Bignum, Real):
      case Pair(Real, Bignum):
      case Pair(Real, Real):{
	    C_Real  B,
	            Ex;

	    R(Left_Op) = mpFloat(R(Left_Op));
	    B = Real_Value(R(Left_Op));

	    if ((B < 0 && Kind(R(Right_Op)) == Real)) {
		R(ans) = OM;
	    } else {
		R(Right_Op) = mpFloat(R(Right_Op));
		Ex = Real_Value(R(Right_Op));
		if (B == 0 && Ex <= 0) {
		    R(ans) = OM;
		} else {
		    R(ans) = New_Real(pow(B, Ex));
		}
	    }
	} break;

      default:
	BAD_ARGS("**", R(Left_Op), R(Right_Op));
    }
    BIN_LEAVE;
}


